"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const client_1 = require("../src/client");
const utils_1 = require("../src/utils");
const ids_json_1 = __importDefault(require("../src/ids.json"));
const web3_js_1 = require("@solana/web3.js");
const serum_1 = require("@project-serum/serum");
const chai_1 = require("chai");
const child_process_1 = require("child_process");
const buffer_layout_1 = require("buffer-layout");
const utils_2 = require("../src/utils");
const dotenv_1 = __importDefault(require("dotenv"));
dotenv_1.default.config();
if (!process.env.AGGREGATOR_PATH) {
    console.info("You have not set the AGGREGATOR_PATH in .env, some tests will fail");
}
const test_utils_1 = require("./test_utils");
console.log = function () { }; // NOTE: Disable all unnecessary logging
let cluster = "devnet";
const client = new client_1.MangoClient();
const clusterIds = ids_json_1.default[cluster];
const mangoProgramId = new web3_js_1.PublicKey(clusterIds.mango_program_id);
const dexProgramId = new web3_js_1.PublicKey(clusterIds.dex_program_id);
let connection = new web3_js_1.Connection(ids_json_1.default.cluster_urls[cluster], 'singleGossip');
let mainnetCluster = "mainnet-beta";
const mainnetClusterIds = ids_json_1.default[mainnetCluster];
const mainnetMangoProgramId = new web3_js_1.PublicKey(mainnetClusterIds.mango_program_id);
const mainnetDexProgramId = new web3_js_1.PublicKey(mainnetClusterIds.dex_program_id);
let mainnetConnection = new web3_js_1.Connection(ids_json_1.default.cluster_urls[mainnetCluster], 'singleGossip');
function chunkOrders(orders, chunkSize) {
    return orders.reduce((resultArray, item, index) => {
        const chunkIndex = Math.floor(index / chunkSize);
        if (!resultArray[chunkIndex]) {
            resultArray[chunkIndex] = [];
        }
        resultArray[chunkIndex].push(item);
        return resultArray;
    }, []);
}
function initAccountsWithBalances(neededBalances, wrappedSol) {
    return __awaiter(this, void 0, void 0, function* () {
        const owner = yield test_utils_1.createWalletAndRequestAirdrop(connection, 5);
        const mangoGroupTokenMappings = yield test_utils_1.createMangoGroupSymbolMappings(connection, mangoGroupIds);
        yield Promise.all(neededBalances.map((x, i) => __awaiter(this, void 0, void 0, function* () {
            if (x > 0) {
                const baseSymbol = mangoGroupSymbols[i];
                yield test_utils_1.createTokenAccountWithBalance(connection, owner, baseSymbol, mangoGroupTokenMappings, clusterIds.faucets, x, wrappedSol);
            }
        })));
        test_utils_1.prettyPrintOwnerKeys(owner, "Account");
        return owner;
    });
}
function requestPriceChange(mangoGroup, requiredPrice, baseSymbol) {
    return __awaiter(this, void 0, void 0, function* () {
        let prices = yield mangoGroup.getPrices(connection);
        while (prices[0].toFixed(2) !== requiredPrice.toFixed(2)) {
            console.info("Running oracle to change price");
            yield performPriceChange(Math.round(requiredPrice * 1e2) / 1e2, baseSymbol.toLowerCase());
            console.info("Finished running oracle to change price");
            try {
                prices = yield mangoGroup.getPrices(connection);
            }
            catch (e) {
                console.info("Error, trying to reset connection");
                connection = new web3_js_1.Connection(ids_json_1.default.cluster_urls[cluster], 'singleGossip');
                prices = yield mangoGroup.getPrices(connection);
            }
        }
        return prices;
    });
}
function performPriceChange(requiredPrice, baseSymbol) {
    return new Promise(function (resolve, _) {
        const priceChangerOracle = child_process_1.spawn('yarn', ['solink', 'oracle', (requiredPrice * 100).toString()], { cwd: process.env.AGGREGATOR_PATH });
        priceChangerOracle.stdout.on("data", data => {
            if (data.includes(`Submit OK {"aggregator":"${baseSymbol}:usd"`)) {
                priceChangerOracle.kill();
                resolve();
            }
        });
    });
}
function cleanOrderBook(mangoGroupSpotMarket) {
    return __awaiter(this, void 0, void 0, function* () {
        console.info("Cleaning order book, this will take a while...");
        const mangoGroup = yield client.getMangoGroup(connection, mangoGroupPk);
        const mangoGroupTokenMappings = yield test_utils_1.createMangoGroupSymbolMappings(connection, mangoGroupIds);
        const { spotMarket, baseSymbol, quoteSymbol } = yield test_utils_1.getSpotMarketDetails(connection, mangoGroupSpotMarket, dexProgramId);
        let bna = yield test_utils_1.getAndDecodeBidsAndAsks(connection, spotMarket);
        let allAsks = [...bna.askOrderBook].map(x => ({ price: x.price, size: x.size })).reverse();
        let allBids = [...bna.bidOrderBook].map(x => ({ price: x.price, size: x.size })).reverse();
        if (allAsks.length || allBids.length) {
            const owner = yield test_utils_1.createWalletAndRequestAirdrop(connection, 5);
            try {
                const marginAccountPk = yield client.initMarginAccount(connection, mangoProgramId, mangoGroup, owner);
                let marginAccount = yield client.getMarginAccount(connection, marginAccountPk, dexProgramId);
                const amountNeededToClearAsks = Math.ceil(allAsks.reduce((acc, ask) => acc + (ask.price * ask.size), 0) + 10);
                yield test_utils_1.createTokenAccountWithBalance(connection, owner, quoteSymbol, mangoGroupTokenMappings, clusterIds.faucets, amountNeededToClearAsks);
                yield test_utils_1.performSingleDepositOrWithdrawal(connection, owner, client, mangoGroup, mangoProgramId, quoteSymbol, mangoGroupTokenMappings, marginAccount, 'deposit', amountNeededToClearAsks);
                const chunkedAsks = chunkOrders(allAsks, 15);
                for (let ask of chunkedAsks) {
                    marginAccount = yield client.getMarginAccount(connection, marginAccountPk, dexProgramId);
                    const price = Math.max(...ask.map((x) => x.price));
                    const size = ask.reduce((a, b) => a + b.size, 0);
                    const roundedSize = Math.round(size * 1e6) / 1e6;
                    console.info(`Buying ${roundedSize} for ${price}`);
                    yield client.placeAndSettle(connection, mangoProgramId, mangoGroup, marginAccount, spotMarket, owner, 'buy', price, roundedSize);
                }
                const amountNeededToClearBids = Math.ceil(allBids.reduce((acc, bid) => acc + (bid.size), 0) + 10);
                yield test_utils_1.createTokenAccountWithBalance(connection, owner, baseSymbol, mangoGroupTokenMappings, clusterIds.faucets, amountNeededToClearBids);
                yield test_utils_1.performSingleDepositOrWithdrawal(connection, owner, client, mangoGroup, mangoProgramId, baseSymbol, mangoGroupTokenMappings, marginAccount, 'deposit', amountNeededToClearBids);
                const chunkedBids = chunkOrders(allBids, 15);
                for (let bid of chunkedBids) {
                    marginAccount = yield client.getMarginAccount(connection, marginAccountPk, dexProgramId);
                    const price = Math.min(...bid.map((x) => x.price));
                    const size = bid.reduce((a, b) => a + b.size, 0);
                    const roundedSize = Math.round(size * 1e6) / 1e6;
                    console.info(`Selling ${roundedSize} for ${price}`);
                    yield client.placeAndSettle(connection, mangoProgramId, mangoGroup, marginAccount, spotMarket, owner, 'sell', price, roundedSize);
                }
                bna = yield test_utils_1.getAndDecodeBidsAndAsks(connection, spotMarket);
                allAsks = [...bna.askOrderBook].map(x => ({ price: x.price, size: x.size }));
                allBids = [...bna.bidOrderBook].map(x => ({ price: x.price, size: x.size }));
                chai_1.expect(allAsks).to.be.empty;
                chai_1.expect(allBids).to.be.empty;
                test_utils_1.prettyPrintOwnerKeys(owner, "Cleaner");
            }
            catch (error) {
                console.info(error);
                throw new Error(`
        Test Error: ${error.message},
        ${test_utils_1.prettyPrintOwnerKeys(owner, "Cleaner")}
      `);
            }
        }
    });
}
function placeNOrdersAfterLimit(mangoGroupSpotMarket, marketIndex, orderQuantityAfter) {
    return __awaiter(this, void 0, void 0, function* () {
        let openOrdersForOwner;
        const orderQuantity = 128; // Max orders
        const buyerOwner = yield test_utils_1.createWalletAndRequestAirdrop(connection, 5);
        test_utils_1.prettyPrintOwnerKeys(buyerOwner, "Buyer");
        const mangoGroup = yield client.getMangoGroup(connection, mangoGroupPk);
        const mangoGroupTokenMappings = yield test_utils_1.createMangoGroupSymbolMappings(connection, mangoGroupIds);
        const buyerMarginAccountPk = yield client.initMarginAccount(connection, mangoProgramId, mangoGroup, buyerOwner);
        let buyerMarginAccount = yield client.getMarginAccount(connection, buyerMarginAccountPk, dexProgramId);
        const { spotMarket, baseSymbol, quoteSymbol } = yield test_utils_1.getSpotMarketDetails(connection, mangoGroupSpotMarket, dexProgramId);
        const [orderSize, orderPrice, _] = yield test_utils_1.getOrderSizeAndPrice(connection, spotMarket, mangoGroupTokenMappings, baseSymbol, quoteSymbol, 'buy');
        const neededQuoteAmount = orderPrice * orderSize;
        const neededQuoteAmountForAllTrades = neededQuoteAmount * orderQuantity;
        yield test_utils_1.createTokenAccountWithBalance(connection, buyerOwner, baseSymbol, mangoGroupTokenMappings, clusterIds.faucets, neededQuoteAmountForAllTrades);
        yield test_utils_1.performSingleDepositOrWithdrawal(connection, buyerOwner, client, mangoGroup, mangoProgramId, baseSymbol, mangoGroupTokenMappings, buyerMarginAccount, 'deposit', neededQuoteAmountForAllTrades);
        yield requestPriceChange(mangoGroup, orderPrice, baseSymbol);
        for (let i = 0; i < orderQuantity; i++) {
            console.info(`Placing a buy order of ${orderSize} ${baseSymbol} for ${orderPrice} ${quoteSymbol} = ~${neededQuoteAmount} ${quoteSymbol} - ${i + 1}/${orderQuantity}`);
            buyerMarginAccount = yield client.getMarginAccount(connection, buyerMarginAccount.publicKey, dexProgramId);
            yield client.placeAndSettle(connection, mangoProgramId, mangoGroup, buyerMarginAccount, spotMarket, buyerOwner, 'buy', orderPrice, orderSize);
        }
        buyerMarginAccount = yield client.getMarginAccount(connection, buyerMarginAccount.publicKey, dexProgramId);
        openOrdersForOwner = yield test_utils_1.getAndDecodeBidsAndAsksForOwner(connection, spotMarket, buyerMarginAccount.openOrdersAccounts[marketIndex]);
        // TODO: this should be a for loop of cancellations
        chai_1.expect(openOrdersForOwner).to.be.an('array').and.to.have.lengthOf(128);
        yield client.cancelOrder(connection, mangoProgramId, mangoGroup, buyerMarginAccount, buyerOwner, spotMarket, openOrdersForOwner[0]);
        buyerMarginAccount = yield client.getMarginAccount(connection, buyerMarginAccount.publicKey, dexProgramId);
        openOrdersForOwner = yield test_utils_1.getAndDecodeBidsAndAsksForOwner(connection, spotMarket, buyerMarginAccount.openOrdersAccounts[marketIndex]);
        chai_1.expect(openOrdersForOwner).to.be.an('array').and.to.have.lengthOf(127);
        for (let i = 0; i < orderQuantityAfter; i++) {
            console.info(`Placing a buy order of ${orderSize} ${baseSymbol} for ${orderPrice} ${quoteSymbol} = ~${neededQuoteAmount} ${quoteSymbol} - ${i + 1}/${orderQuantityAfter}`);
            buyerMarginAccount = yield client.getMarginAccount(connection, buyerMarginAccount.publicKey, dexProgramId);
            yield client.placeAndSettle(connection, mangoProgramId, mangoGroup, buyerMarginAccount, spotMarket, buyerOwner, 'buy', orderPrice, orderSize);
        }
        chai_1.expect(openOrdersForOwner).to.be.an('array').and.to.have.lengthOf(128);
    });
}
function stressTestMatchOrder(mangoGroupSpotMarket, orderQuantity) {
    return __awaiter(this, void 0, void 0, function* () {
        let bna, allAsks, allBids;
        const sellerOwner = yield test_utils_1.createWalletAndRequestAirdrop(connection, 5);
        test_utils_1.prettyPrintOwnerKeys(sellerOwner, "Seller");
        const mangoGroup = yield client.getMangoGroup(connection, mangoGroupPk);
        const mangoGroupTokenMappings = yield test_utils_1.createMangoGroupSymbolMappings(connection, mangoGroupIds);
        const sellerMarginAccountPk = yield client.initMarginAccount(connection, mangoProgramId, mangoGroup, sellerOwner);
        let sellerMarginAccount = yield client.getMarginAccount(connection, sellerMarginAccountPk, dexProgramId);
        const { spotMarket, baseSymbol, quoteSymbol } = yield test_utils_1.getSpotMarketDetails(connection, mangoGroupSpotMarket, dexProgramId);
        const [orderSize, orderPrice, _] = yield test_utils_1.getOrderSizeAndPrice(connection, spotMarket, mangoGroupTokenMappings, baseSymbol, quoteSymbol, 'sell');
        const neededQuoteAmount = orderPrice * orderSize;
        const neededBaseAmountForAllTrades = orderSize * orderQuantity;
        const neededQuoteAmountForAllTrades = neededQuoteAmount * orderQuantity;
        yield test_utils_1.createTokenAccountWithBalance(connection, sellerOwner, baseSymbol, mangoGroupTokenMappings, clusterIds.faucets, neededBaseAmountForAllTrades);
        yield test_utils_1.performSingleDepositOrWithdrawal(connection, sellerOwner, client, mangoGroup, mangoProgramId, baseSymbol, mangoGroupTokenMappings, sellerMarginAccount, 'deposit', neededBaseAmountForAllTrades);
        for (let i = 0; i < orderQuantity; i++) {
            console.info(`Placing a sell order of ${orderSize} ${baseSymbol} for ${orderPrice} ${quoteSymbol} = ~${neededQuoteAmount} USD - ${i + 1}/${orderQuantity}`);
            sellerMarginAccount = yield client.getMarginAccount(connection, sellerMarginAccountPk, dexProgramId);
            yield client.placeAndSettle(connection, mangoProgramId, mangoGroup, sellerMarginAccount, spotMarket, sellerOwner, 'sell', orderPrice, orderSize);
        }
        bna = yield test_utils_1.getAndDecodeBidsAndAsks(connection, spotMarket);
        allAsks = [...bna.askOrderBook].map(x => ({ price: x.price, size: x.size }));
        allBids = [...bna.bidOrderBook].map(x => ({ price: x.price, size: x.size }));
        const buyerOwner = yield test_utils_1.createWalletAndRequestAirdrop(connection, 5);
        test_utils_1.prettyPrintOwnerKeys(buyerOwner, "Buyer");
        const buyerMarginAccountPk = yield client.initMarginAccount(connection, mangoProgramId, mangoGroup, buyerOwner);
        const buyerMarginAccount = yield client.getMarginAccount(connection, buyerMarginAccountPk, dexProgramId);
        yield test_utils_1.createTokenAccountWithBalance(connection, buyerOwner, quoteSymbol, mangoGroupTokenMappings, clusterIds.faucets, neededQuoteAmountForAllTrades);
        yield test_utils_1.performSingleDepositOrWithdrawal(connection, buyerOwner, client, mangoGroup, mangoProgramId, quoteSymbol, mangoGroupTokenMappings, buyerMarginAccount, 'deposit', neededQuoteAmountForAllTrades);
        console.info(`Placing a buy order of ${neededBaseAmountForAllTrades} ${baseSymbol} for ${orderPrice} ${quoteSymbol} = ~${neededQuoteAmountForAllTrades} ${quoteSymbol}`);
        const buyTxHash = yield client.placeAndSettle(connection, mangoProgramId, mangoGroup, buyerMarginAccount, spotMarket, buyerOwner, 'buy', orderPrice, neededBaseAmountForAllTrades);
        console.info("buyTxHash:", buyTxHash);
        yield connection.confirmTransaction(buyTxHash, 'finalized');
        const buyConfirmedTx = yield connection.getConfirmedTransaction(buyTxHash);
        const buyTxLogInfo = test_utils_1.extractInfoFromLogs(buyConfirmedTx);
        console.info("Buy txLogInfo:", buyTxLogInfo);
        bna = yield test_utils_1.getAndDecodeBidsAndAsks(connection, spotMarket);
        allAsks = [...bna.askOrderBook].map(x => ({ price: x.price, size: x.size }));
        allBids = [...bna.bidOrderBook].map(x => ({ price: x.price, size: x.size }));
        chai_1.expect(allAsks).to.be.empty;
        chai_1.expect(allBids).to.be.empty;
    });
}
function stressTestLiquidation(params) {
    return __awaiter(this, void 0, void 0, function* () {
        const { mangoGroupSpotMarket, orderQuantity = 1, customLiqeeOwner = null, shouldPartialLiquidate = false, shouldCreateNewLiqor = true, shouldFinishLiquidationInTest = true, customOrderPrice = 0, customOrderSize = 0, leverageCoefficient = 15, matchLeveragedOrder = false, side = 'buy' } = params;
        console.info("shouldCreateNewLiqor:", shouldCreateNewLiqor);
        console.info("orderQuantity:", orderQuantity);
        let bna, allAsks, allBids, prices;
        const liqeeOwner = customLiqeeOwner || (yield test_utils_1.createWalletAndRequestAirdrop(connection, 5));
        test_utils_1.prettyPrintOwnerKeys(liqeeOwner, "Liqee");
        const mangoGroup = yield client.getMangoGroup(connection, mangoGroupPk);
        const mangoGroupTokenMappings = yield test_utils_1.createMangoGroupSymbolMappings(connection, mangoGroupIds);
        const liqeeMarginAccountPk = (customLiqeeOwner) ? (yield client.getMarginAccountsForOwner(connection, mangoProgramId, mangoGroup, liqeeOwner))[0].publicKey : yield client.initMarginAccount(connection, mangoProgramId, mangoGroup, liqeeOwner);
        let liqeeMarginAccount = yield client.getMarginAccount(connection, liqeeMarginAccountPk, dexProgramId);
        const { spotMarket, baseSymbol, quoteSymbol } = yield test_utils_1.getSpotMarketDetails(connection, mangoGroupSpotMarket, dexProgramId);
        const baseSymbolIndex = mangoGroupSymbols.findIndex(x => x === baseSymbol);
        const quoteSymbolIndex = mangoGroupSymbols.findIndex(x => x === quoteSymbol);
        const [orderSize, orderPrice, _] = yield test_utils_1.getOrderSizeAndPrice(connection, spotMarket, mangoGroupTokenMappings, baseSymbol, quoteSymbol, side);
        const finalOrderPrice = customOrderPrice || orderPrice;
        const finalOrderSize = customOrderSize || orderSize;
        const neededQuoteAmount = finalOrderPrice * finalOrderSize;
        const neededBaseAmountForAllTrades = finalOrderSize * orderQuantity;
        const neededQuoteAmountForAllTrades = neededQuoteAmount * orderQuantity;
        const neededCollateralTokenSymbol = (side === 'buy') ? baseSymbol : quoteSymbol;
        const neededAmountForAllTrades = (side === 'buy') ? neededBaseAmountForAllTrades : neededQuoteAmountForAllTrades;
        yield test_utils_1.createTokenAccountWithBalance(connection, liqeeOwner, neededCollateralTokenSymbol, mangoGroupTokenMappings, clusterIds.faucets, neededAmountForAllTrades);
        yield test_utils_1.performSingleDepositOrWithdrawal(connection, liqeeOwner, client, mangoGroup, mangoProgramId, neededCollateralTokenSymbol, mangoGroupTokenMappings, liqeeMarginAccount, 'deposit', neededAmountForAllTrades);
        prices = yield requestPriceChange(mangoGroup, finalOrderPrice, baseSymbol);
        for (let i = 0; i < orderQuantity; i++) {
            console.info(`Placing a ${side} order of ${finalOrderSize} ${baseSymbol} for ${finalOrderPrice} ${quoteSymbol} = ~${neededQuoteAmount} ${quoteSymbol} - ${i + 1}/${orderQuantity}`);
            liqeeMarginAccount = yield client.getMarginAccount(connection, liqeeMarginAccountPk, dexProgramId);
            console.info("Deposits init:", liqeeMarginAccount.getAssets(mangoGroup));
            console.info("Assets init:", liqeeMarginAccount.getAssets(mangoGroup));
            console.info("Liabs init:", liqeeMarginAccount.getLiabs(mangoGroup));
            yield client.placeAndSettle(connection, mangoProgramId, mangoGroup, liqeeMarginAccount, spotMarket, liqeeOwner, side, finalOrderPrice, finalOrderSize);
            console.info("Sleeep!!!");
            yield utils_2.sleep(5000);
            liqeeMarginAccount = yield client.getMarginAccount(connection, liqeeMarginAccountPk, dexProgramId);
            console.info("Deposits:", liqeeMarginAccount.getAssets(mangoGroup));
            console.info("Assets:", liqeeMarginAccount.getAssets(mangoGroup));
            console.info("Assets Val:", liqeeMarginAccount.getAssetsVal(mangoGroup, prices));
            console.info("Liabs:", liqeeMarginAccount.getLiabs(mangoGroup));
        }
        if (matchLeveragedOrder)
            yield cleanOrderBook(mangoGroupSpotMarket);
        console.info("Sleeep!!!");
        yield utils_2.sleep(10000);
        liqeeMarginAccount = yield client.getMarginAccount(connection, liqeeMarginAccountPk, dexProgramId);
        console.info("collRatio before price change:", liqeeMarginAccount.getCollateralRatio(mangoGroup, prices));
        console.info("Assets before:", liqeeMarginAccount.getAssets(mangoGroup));
        console.info("Assets Val before:", liqeeMarginAccount.getAssetsVal(mangoGroup, prices));
        console.info("Deposits before:", liqeeMarginAccount.getDeposits(mangoGroup));
        console.info("Liabs before:", liqeeMarginAccount.getLiabs(mangoGroup));
        console.info(prices);
        const adjustedPrice = (side === 'buy') ? finalOrderPrice / leverageCoefficient : finalOrderPrice * leverageCoefficient;
        prices = yield requestPriceChange(mangoGroup, adjustedPrice, baseSymbol);
        liqeeMarginAccount = yield client.getMarginAccount(connection, liqeeMarginAccountPk, dexProgramId);
        console.info("collRatio after price change:", liqeeMarginAccount.getCollateralRatio(mangoGroup, prices));
        console.info("Assets after:", liqeeMarginAccount.getAssets(mangoGroup));
        console.info("Assets Val after:", liqeeMarginAccount.getAssetsVal(mangoGroup, prices));
        console.info("Deposits after:", liqeeMarginAccount.getDeposits(mangoGroup));
        console.info("Liabs after:", liqeeMarginAccount.getLiabs(mangoGroup));
        console.info(prices);
        let liqorOwner = new web3_js_1.Account();
        if (shouldCreateNewLiqor) {
            liqorOwner = yield test_utils_1.createWalletAndRequestAirdrop(connection, 5);
            test_utils_1.prettyPrintOwnerKeys(liqorOwner, "Liqor");
            for (let mangoGroupSymbol of mangoGroupSymbols) {
                const requiredBalance = (mangoGroupSymbol === quoteSymbol) ? neededQuoteAmountForAllTrades : 0;
                yield test_utils_1.createTokenAccountWithBalance(connection, liqorOwner, mangoGroupSymbol, mangoGroupTokenMappings, clusterIds.faucets, requiredBalance);
            }
            if (shouldFinishLiquidationInTest) {
                const tokenWallets = (yield Promise.all(mangoGroup.tokens.map((mint) => utils_1.findLargestTokenAccountForOwner(connection, liqorOwner.publicKey, mint).then((response) => response.publicKey))));
                let liquidationTxHash;
                if (shouldPartialLiquidate) {
                    const forceCancelTx = yield client.forceCancelOrders(connection, mangoProgramId, mangoGroup, liqeeMarginAccount, liqorOwner, spotMarket, 5);
                    console.log('forceCancelTxHash', forceCancelTx);
                    liquidationTxHash = yield client.partialLiquidate(connection, mangoProgramId, mangoGroup, liqeeMarginAccount, liqorOwner, tokenWallets[quoteSymbolIndex], tokenWallets[baseSymbolIndex], quoteSymbolIndex, baseSymbolIndex, neededQuoteAmountForAllTrades);
                }
                else {
                    const depositQuantities = new Array(tokenWallets.length).fill(0);
                    depositQuantities[quoteSymbolIndex] = neededQuoteAmountForAllTrades;
                    liquidationTxHash = yield client.liquidate(connection, mangoProgramId, mangoGroup, liqeeMarginAccount, liqorOwner, tokenWallets, depositQuantities);
                }
                console.info("liquidationTxHash:", liquidationTxHash);
                yield connection.confirmTransaction(liquidationTxHash, 'finalized');
                const liquidationConfirmedTx = yield connection.getConfirmedTransaction(liquidationTxHash);
                const liquidationTxLogInfo = test_utils_1.extractInfoFromLogs(liquidationConfirmedTx);
                console.info("Liquidation txLogInfo:", liquidationTxLogInfo);
            }
        }
        return { liqeeOwner, liqorOwner };
    });
}
const mangoGroupName = 'BTC_ETH_SOL_SRM_USDC';
const mangoGroupSymbols = mangoGroupName.split('_');
const mangoGroupIds = clusterIds.mango_groups[mangoGroupName];
const mangoGroupSpotMarkets = Object.entries(mangoGroupIds.spot_market_symbols);
const mangoGroupPk = new web3_js_1.PublicKey(mangoGroupIds.mango_group_pk);
const mangoGroupSpotMarket = mangoGroupSpotMarkets[0]; //BTC/USDC
// const mangoGroupSpotMarket = mangoGroupSpotMarkets[1]; //ETH/USDC
// const mangoGroupSpotMarket = mangoGroupSpotMarkets[2]; //SOL/USDC
// const mangoGroupSpotMarket = mangoGroupSpotMarkets[3]; //SRM/USDC
describe('stress test order limits', () => __awaiter(void 0, void 0, void 0, function* () {
    before(() => __awaiter(void 0, void 0, void 0, function* () {
        yield cleanOrderBook(mangoGroupSpotMarket);
    }));
    it('should be able to place 129th order after cancelling one', () => __awaiter(void 0, void 0, void 0, function* () {
        yield placeNOrdersAfterLimit(mangoGroupSpotMarket, 0, 1);
    }));
}));
describe('stress testing matching orders', () => __awaiter(void 0, void 0, void 0, function* () {
    before(() => __awaiter(void 0, void 0, void 0, function* () {
        yield cleanOrderBook(mangoGroupSpotMarket);
    }));
    it('should match 1 order at a single price', () => __awaiter(void 0, void 0, void 0, function* () {
        yield stressTestMatchOrder(mangoGroupSpotMarket, 1);
    }));
    it('should match 10 orders at a single price', () => __awaiter(void 0, void 0, void 0, function* () {
        yield stressTestMatchOrder(mangoGroupSpotMarket, 10);
    }));
    it('should match 20 orders at a single price', () => __awaiter(void 0, void 0, void 0, function* () {
        yield stressTestMatchOrder(mangoGroupSpotMarket, 20);
    }));
    it('should match 25 orders at a single price', () => __awaiter(void 0, void 0, void 0, function* () {
        yield stressTestMatchOrder(mangoGroupSpotMarket, 25);
    }));
}));
describe('stress testing liquidation', () => __awaiter(void 0, void 0, void 0, function* () {
    before(() => __awaiter(void 0, void 0, void 0, function* () {
        yield cleanOrderBook(mangoGroupSpotMarket);
    }));
    it('should liquidate an account with 1 open order', () => __awaiter(void 0, void 0, void 0, function* () {
        yield stressTestLiquidation({ mangoGroupSpotMarket, orderQuantity: 1 });
    }));
    it('should liquidate an account with 10 open orders', () => __awaiter(void 0, void 0, void 0, function* () {
        yield stressTestLiquidation({ mangoGroupSpotMarket, orderQuantity: 10 });
    }));
    it('should liquidate an account with 20 open orders', () => __awaiter(void 0, void 0, void 0, function* () {
        yield stressTestLiquidation({ mangoGroupSpotMarket, orderQuantity: 20 });
    }));
    it('should liquidate an account with 25 open orders', () => __awaiter(void 0, void 0, void 0, function* () {
        yield stressTestLiquidation({ mangoGroupSpotMarket, orderQuantity: 25 });
    }));
    it('should liquidate an account with 128 open orders', () => __awaiter(void 0, void 0, void 0, function* () {
        yield stressTestLiquidation({ mangoGroupSpotMarket, orderQuantity: 128 });
    }));
}));
describe('stress testing partial liquidation', () => __awaiter(void 0, void 0, void 0, function* () {
    // before(async () => {
    //   await cleanOrderBook(mangoGroupSpotMarket);
    // });
    it('should partially liquidate an account with 1 open order', () => __awaiter(void 0, void 0, void 0, function* () {
        yield stressTestLiquidation({ mangoGroupSpotMarket, orderQuantity: 1, shouldPartialLiquidate: true });
    }));
    it('should partially liquidate an account with 10 open orders', () => __awaiter(void 0, void 0, void 0, function* () {
        yield stressTestLiquidation({ mangoGroupSpotMarket, orderQuantity: 10, shouldPartialLiquidate: true });
    }));
    it('should partially liquidate an account with 20 open orders', () => __awaiter(void 0, void 0, void 0, function* () {
        yield stressTestLiquidation({ mangoGroupSpotMarket, orderQuantity: 20, shouldPartialLiquidate: true });
    }));
    it('should partially liquidate an account with 25 open orders', () => __awaiter(void 0, void 0, void 0, function* () {
        yield stressTestLiquidation({ mangoGroupSpotMarket, orderQuantity: 25, shouldPartialLiquidate: true });
    }));
    it('should partially liquidate an account with 128 open orders', () => __awaiter(void 0, void 0, void 0, function* () {
        yield stressTestLiquidation({ mangoGroupSpotMarket, orderQuantity: 128, shouldPartialLiquidate: true });
    }));
    it('should test socialized loss with 1 borrows', () => __awaiter(void 0, void 0, void 0, function* () {
        const mangoGroupSpotMarketBTC = mangoGroupSpotMarkets[0]; //BTC/USDC
        yield stressTestLiquidation({ mangoGroupSpotMarket: mangoGroupSpotMarketBTC, shouldPartialLiquidate: true, shouldFinishLiquidationInTest: true, customOrderPrice: 20, customOrderSize: 1, matchLeveragedOrder: false, shouldCreateNewLiqor: true, leverageCoefficient: 30 });
    }));
    it('should test socialized loss with 4 borrows', () => __awaiter(void 0, void 0, void 0, function* () {
        const mangoGroupSpotMarketETH = mangoGroupSpotMarkets[1]; //ETH/USDC
        const { liqeeOwner } = yield stressTestLiquidation({ mangoGroupSpotMarket: mangoGroupSpotMarketETH, shouldPartialLiquidate: true, shouldFinishLiquidationInTest: false, customOrderPrice: 10, customOrderSize: 1, shouldCreateNewLiqor: false });
        const mangoGroupSpotMarketSOL = mangoGroupSpotMarkets[2]; //SOL/USDC
        yield stressTestLiquidation({ mangoGroupSpotMarket: mangoGroupSpotMarketSOL, customLiqeeOwner: liqeeOwner, shouldPartialLiquidate: true, shouldFinishLiquidationInTest: false, customOrderPrice: 10, customOrderSize: 1, shouldCreateNewLiqor: false });
        const mangoGroupSpotMarketSRM = mangoGroupSpotMarkets[3]; //SRM/USDC
        yield stressTestLiquidation({ mangoGroupSpotMarket: mangoGroupSpotMarketSRM, customLiqeeOwner: liqeeOwner, shouldPartialLiquidate: true, shouldFinishLiquidationInTest: false, customOrderPrice: 10, customOrderSize: 1, shouldCreateNewLiqor: false });
        const mangoGroupSpotMarketBTC = mangoGroupSpotMarkets[0]; //BTC/USDC
        yield stressTestLiquidation({ mangoGroupSpotMarket: mangoGroupSpotMarketBTC, customLiqeeOwner: liqeeOwner, shouldPartialLiquidate: true, shouldFinishLiquidationInTest: true, customOrderPrice: 10, customOrderSize: 1, matchLeveragedOrder: true, shouldCreateNewLiqor: true });
    }));
}));
describe('create various liquidation opportunities', () => __awaiter(void 0, void 0, void 0, function* () {
    it('should create a liquidation opportunity', () => __awaiter(void 0, void 0, void 0, function* () {
        const { liqeeOwner, liqorOwner } = yield stressTestLiquidation({ mangoGroupSpotMarket: mangoGroupSpotMarket, shouldPartialLiquidate: true, shouldFinishLiquidationInTest: false, customOrderPrice: 20, customOrderSize: 1, side: 'sell' });
        // const mangoGroupTokenMappings = await createMangoGroupSymbolMappings(connection, mangoGroupIds);
        // const quoteCurrencyAccountPk = await createTokenAccountWithBalance(connection, liqorOwner, 'USDC', mangoGroupTokenMappings, clusterIds.faucets, 0);
        // const baseCurrencyAmounts: number[] = [1, 2, 10, 100];
        // await Promise.all(mangoGroupSpotMarkets.map(async (mangoGroupSpotMarket: any, index: number) => {
        //   const { spotMarket, baseSymbol, minSize, minPrice } = await getSpotMarketDetails(connection, mangoGroupSpotMarket, dexProgramId);
        //   console.info(baseSymbol);
        //   const baseCurrencyAccountPk = await createTokenAccountWithBalance(connection, liqorOwner, baseSymbol, mangoGroupTokenMappings, clusterIds.faucets, baseCurrencyAmounts[index]);
        //   const side = 'sell';
        //   if (!baseCurrencyAccountPk || !quoteCurrencyAccountPk) throw Error('Missing the necessary token accounts');
        //   await placeOrderUsingSerumDex(connection, liqorOwner, spotMarket, baseCurrencyAccountPk, quoteCurrencyAccountPk, { side, size: minSize, price: minPrice });
        //   // const openOrdersAccounts = await spotMarket.findOpenOrdersAccountsForOwner(connection, owner.publicKey);
        //   // const ordersForOwner = await getAndDecodeBidsAndAsksForOwner(connection, spotMarket, openOrdersAccounts[0]);
        //   // await cancelOrdersUsingSerumDex(connection, owner, spotMarket, ordersForOwner);
        //   const allOpenOrdersAccounts = await OpenOrders.findForOwner(connection, liqorOwner.publicKey, dexProgramId)
        //   console.info("All oo accs:", allOpenOrdersAccounts.length);
        // }));
    }));
}));
describe('create accounts for testing', () => __awaiter(void 0, void 0, void 0, function* () {
    it('should create an account with test money', () => __awaiter(void 0, void 0, void 0, function* () {
        const buyerOwner = yield initAccountsWithBalances([100, 100, 500, 1000, 100000], false);
    }));
    it('should fund mango group', () => __awaiter(void 0, void 0, void 0, function* () {
        const buyerOwner = yield initAccountsWithBalances([100, 100, 500, 1000, 100000], false);
        const amounts = [100, 100, 500, 1000, 100000];
        const symbols = ['BTC', 'ETH', 'SOL', 'SRM', 'USDC'];
        const mangoGroup = yield client.getMangoGroup(connection, mangoGroupPk);
        const mangoGroupTokenMappings = yield test_utils_1.createMangoGroupSymbolMappings(connection, mangoGroupIds);
        const buyerMarginAccountPk = yield client.initMarginAccount(connection, mangoProgramId, mangoGroup, buyerOwner);
        const buyerMarginAccount = yield client.getMarginAccount(connection, buyerMarginAccountPk, dexProgramId);
        for (let i = 0; i < symbols.length; i++) {
            yield test_utils_1.createTokenAccountWithBalance(connection, buyerOwner, symbols[i], mangoGroupTokenMappings, clusterIds.faucets, amounts[i]);
            yield test_utils_1.performSingleDepositOrWithdrawal(connection, buyerOwner, client, mangoGroup, mangoProgramId, symbols[i], mangoGroupTokenMappings, buyerMarginAccount, 'deposit', amounts[i]);
        }
    }));
    it('should create an account with initialised openOrders', () => __awaiter(void 0, void 0, void 0, function* () {
        const owner = yield test_utils_1.createWalletAndRequestAirdrop(connection, 5);
        const mangoGroupTokenMappings = yield test_utils_1.createMangoGroupSymbolMappings(connection, mangoGroupIds);
        const quoteCurrencyAccountPk = yield test_utils_1.createTokenAccountWithBalance(connection, owner, 'USDC', mangoGroupTokenMappings, clusterIds.faucets, 0);
        yield Promise.all(mangoGroupSpotMarkets.map((mangoGroupSpotMarket) => __awaiter(void 0, void 0, void 0, function* () {
            const { spotMarket, baseSymbol, minSize, minPrice } = yield test_utils_1.getSpotMarketDetails(connection, mangoGroupSpotMarket, dexProgramId);
            const baseCurrencyAccountPk = yield test_utils_1.createTokenAccountWithBalance(connection, owner, baseSymbol, mangoGroupTokenMappings, clusterIds.faucets, minSize);
            const side = 'sell';
            if (!baseCurrencyAccountPk || !quoteCurrencyAccountPk)
                throw Error('Missing the necessary token accounts');
            yield test_utils_1.placeOrderUsingSerumDex(connection, owner, spotMarket, baseCurrencyAccountPk, quoteCurrencyAccountPk, { side, size: minSize, price: minPrice });
            // const openOrdersAccounts = await spotMarket.findOpenOrdersAccountsForOwner(connection, owner.publicKey);
            // const ordersForOwner = await getAndDecodeBidsAndAsksForOwner(connection, spotMarket, openOrdersAccounts[0]);
            // await cancelOrdersUsingSerumDex(connection, owner, spotMarket, ordersForOwner);
        })));
    }));
}));
describe('log stuff', () => __awaiter(void 0, void 0, void 0, function* () {
    // NOTE: This part of tests is used to test and log random things
    it('should log token decimals', () => __awaiter(void 0, void 0, void 0, function* () {
        const MINT_LAYOUT = buffer_layout_1.struct([buffer_layout_1.blob(44), buffer_layout_1.u8('decimals'), buffer_layout_1.blob(37)]);
        const mainnetTokensToTest = [
            ['BTC', '9n4nbM75f5Ui33ZbPYXn59EwSgE8CGsHtAeTH5YFeJ9E'],
            ['ETH', '2FPyTwcZLUg1MDrwsyoP4D6s1tM7hAkHYRjkNb5w6Pxk'],
            ['SOL', 'So11111111111111111111111111111111111111112'],
            ['SRM', 'SRMuApVNdxXokk5GT7XD5cUUgXMBCoAz2LHeuAoKWRt'],
            ['USDT', 'Es9vMFrzaCERmJfrF4H2FYD4KCoNkY11McCe8BenwNYB'],
            ['USDC', 'EPjFWdd5AufqSSqeM2qN1xzybapC8G4wEGGkZwyTDt1v'],
        ];
        const devnetTokensToTest = [
            ['BTC', 'bypQzRBaSDWiKhoAw3hNkf35eF3z3AZCU8Sxks6mTPP'],
            ['ETH', 'ErWGBLBQMwdyC4H3MR8ef6pFK6gyHAyBxy4o1mHoqKzm'],
            ['SOL', 'So11111111111111111111111111111111111111112'],
            ['SRM', '9FbAMDvXqNjPqZSYt4EWTguJuDrGkfvwr3gSFpiSbX9S'],
            ['USDT', '7KBVenLz5WNH4PA5MdGkJNpDDyNKnBQTwnz1UqJv9GUm'],
            ['USDC', 'H6hy7Ykzc43EuGivv7VVuUKNpKgUoFAfUY3wdPr4UyRX'],
        ];
        for (let [tokenName, tokenMint] of mainnetTokensToTest) {
            const data = yield mainnetConnection.getAccountInfo(new web3_js_1.PublicKey(tokenMint));
            const info = MINT_LAYOUT.decode(data.data);
            console.info(`Mainnet ${tokenName} decimals: ${info.decimals}`);
        }
        for (let [tokenName, tokenMint] of devnetTokensToTest) {
            const data = yield connection.getAccountInfo(new web3_js_1.PublicKey(tokenMint));
            const info = MINT_LAYOUT.decode(data.data);
            console.info(`Devnet ${tokenName} decimals: ${info.decimals}`);
        }
        ;
    }));
    it('should log lotSizes', () => __awaiter(void 0, void 0, void 0, function* () {
        const mainnetSpotMarketsToTest = [
            ['BTC/USDC', 'A8YFbxQYFVqKZaoYJLLUVcQiWP7G2MeEgW5wsAQgMvFw'],
            ['BTC/USDT', 'C1EuT9VokAKLiW7i2ASnZUvxDoKuKkCpDDeNxAptuNe4'],
            ['ETH/USDC', '4tSvZvnbyzHXLMTiFonMyxZoHmFqau1XArcRCVHLZ5gX'],
            ['SOL/USDC', '9wFFyRfZBsuAha4YcuxcXLKwMxJR43S7fPfQLusDBzvT'],
            ['SRM/USDC', 'ByRys5tuUWDgL73G8JBAEfkdFf8JWBzPBDHsBVQ5vbQA'],
        ];
        const devnetSpotMarketsToTest = [
            ['BTC/USDC', 'BCqDfFd119UyNEC2HavKdy3F4qhy6EMGirSurNWKgioW'],
            ['ETH/USDC', 'AfB75DQs1E2VoUAMRorUxAz68b18kWZ1uqQuRibGk212'],
            ['SOL/USDC', '6vZd6Ghwkuzpbp7qNzBuRkhcfA9H3S7BJ2LCWSYrjfzo'],
            ['SRM/USDC', '6rRnXBLGzcD5v1q4NfWWZQdgBfqzEuD3g4GqDWVU8yhH'],
        ];
        for (let [spotMarketName, spotMarketAddress] of mainnetSpotMarketsToTest) {
            const spotMarket = yield serum_1.Market.load(mainnetConnection, new web3_js_1.PublicKey(spotMarketAddress), { skipPreflight: true, commitment: 'singleGossip' }, mainnetDexProgramId);
            console.info(`Mainnet ${spotMarketName} base/quote lotSizes: ${spotMarket['_decoded'].baseLotSize.toString()}/${spotMarket['_decoded'].quoteLotSize.toString()}`);
            console.info(`Mainnet ${spotMarketName} baseSizeNumberToLots: ${spotMarket.baseSizeNumberToLots(1)}`);
            console.info(`Mainnet ${spotMarketName} priceNumberToLots: ${spotMarket.priceNumberToLots(1)}`);
        }
        for (let [spotMarketName, spotMarketAddress] of devnetSpotMarketsToTest) {
            const spotMarket = yield serum_1.Market.load(connection, new web3_js_1.PublicKey(spotMarketAddress), { skipPreflight: true, commitment: 'singleGossip' }, dexProgramId);
            console.info(`Devnet ${spotMarketName} base/quote lotSizes: ${spotMarket['_decoded'].baseLotSize.toString()}/${spotMarket['_decoded'].quoteLotSize.toString()}`);
            console.info(`Devnet ${spotMarketName} baseSizeNumberToLots: ${spotMarket.baseSizeNumberToLots(1)}`);
            console.info(`Devnet ${spotMarketName} priceNumberToLots: ${spotMarket.priceNumberToLots(1)}`);
        }
    }));
    it('should log orderbook for spotmarket', () => __awaiter(void 0, void 0, void 0, function* () {
        const { spotMarket } = yield test_utils_1.getSpotMarketDetails(connection, mangoGroupSpotMarket, dexProgramId);
        let bna = yield test_utils_1.getAndDecodeBidsAndAsks(connection, spotMarket);
        let allAsks = [...bna.askOrderBook].map(x => ({ price: x.price, size: x.size })).reverse();
        let allBids = [...bna.bidOrderBook].map(x => ({ price: x.price, size: x.size })).reverse();
        console.info("=== allAsks ===");
        console.info(allAsks);
        console.info("=== allBids ===");
        console.info(allBids);
    }));
    it('should log margin account of private key', () => __awaiter(void 0, void 0, void 0, function* () {
        const ownerKey = [252, 0, 132, 118, 116, 9, 142, 85, 38, 150, 113, 82, 117, 172, 107, 37, 200, 103, 20, 206, 153, 172, 239, 151, 251, 175, 208, 119, 89, 164, 50, 4, 85, 244, 218, 137, 21, 123, 226, 241, 53, 80, 95, 8, 194, 128, 195, 133, 108, 79, 71, 175, 75, 177, 35, 99, 181, 251, 84, 107, 1, 154, 104, 105];
        const owner = new web3_js_1.Account(ownerKey);
        const mangoGroup = yield client.getMangoGroup(connection, mangoGroupPk);
        const marginAccount = yield client.getMarginAccountsForOwner(connection, mangoProgramId, mangoGroup, owner);
        console.info(marginAccount[0].publicKey.toString());
    }));
}));
//# sourceMappingURL=Stateless.test.js.map