"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.prettyPrintOwnerKeys = exports.extractInfoFromLogs = exports.getOrderSizeAndPrice = exports.getBidOrAskPriceEdge = exports.getAndDecodeBidsAndAsksForOwner = exports.getAndDecodeBidsAndAsks = exports.performSingleDepositOrWithdrawal = exports.createTokenAccountWithBalance = exports.cancelOrdersUsingSerumDex = exports.placeOrderUsingSerumDex = exports.getMinSizeAndPriceForMarket = exports.airdropSol = exports.airdropTokens = exports.buildAirdropTokensIx = exports.createMangoGroupSymbolMappings = exports.getSpotMarketDetails = exports.createWalletAndRequestAirdrop = exports.createTokenAccount = exports.createWrappedNativeAccount = exports.createTokenAccountInstrs = exports._sendTransaction = void 0;
const web3_js_1 = require("@solana/web3.js");
const serum_1 = require("@project-serum/serum");
const common_1 = require("@project-serum/common");
const spl_token_1 = require("@solana/spl-token");
const utils_1 = require("../src/utils");
console.log = function () { }; // NOTE: Disable all unnecessary logging
const FAUCET_PROGRAM_ID = new web3_js_1.PublicKey("4bXpkKSV8swHSnwqtzuboGPaPDeEgAn4Vt8GfarV5rZt");
const getPDA = () => {
    return web3_js_1.PublicKey.findProgramAddress([Buffer.from("faucet")], FAUCET_PROGRAM_ID);
};
function _sendTransaction(connection, transaction, signers) {
    return __awaiter(this, void 0, void 0, function* () {
        yield utils_1.sleep(1000);
        const signature = yield connection.sendTransaction(transaction, signers);
        try {
            yield connection.confirmTransaction(signature);
        }
        catch (e) {
            console.info("Error while confirming, trying again");
            yield connection.confirmTransaction(signature);
        }
        return signature;
    });
}
exports._sendTransaction = _sendTransaction;
function createTokenAccountInstrs(connection, newAccountPubkey, mint, ownerPk, lamports) {
    return __awaiter(this, void 0, void 0, function* () {
        if (lamports === undefined)
            lamports = yield connection.getMinimumBalanceForRentExemption(165);
        return [
            web3_js_1.SystemProgram.createAccount({
                fromPubkey: ownerPk,
                newAccountPubkey,
                space: 165,
                lamports,
                programId: serum_1.TokenInstructions.TOKEN_PROGRAM_ID,
            }),
            serum_1.TokenInstructions.initializeAccount({
                account: newAccountPubkey,
                mint,
                owner: ownerPk,
            }),
        ];
    });
}
exports.createTokenAccountInstrs = createTokenAccountInstrs;
function createWrappedNativeAccount(connection, owner, amount) {
    return __awaiter(this, void 0, void 0, function* () {
        // Allocate memory for the account
        const balanceNeeded = yield connection.getMinimumBalanceForRentExemption(165);
        const newAccount = new web3_js_1.Account();
        const tx = new web3_js_1.Transaction();
        tx.add(web3_js_1.SystemProgram.createAccount({
            fromPubkey: owner.publicKey,
            newAccountPubkey: newAccount.publicKey,
            lamports: balanceNeeded,
            space: 165,
            programId: serum_1.TokenInstructions.TOKEN_PROGRAM_ID,
        })); // Send lamports to it (these will be wrapped into native tokens by the token program)
        tx.add(web3_js_1.SystemProgram.transfer({
            fromPubkey: owner.publicKey,
            toPubkey: newAccount.publicKey,
            lamports: amount
        })); // Assign the new account to the native token mint.
        // the account will be initialized with a balance equal to the native token balance.
        // (i.e. amount)
        tx.add(serum_1.TokenInstructions.initializeAccount({
            account: newAccount.publicKey,
            mint: spl_token_1.NATIVE_MINT,
            owner: owner.publicKey,
        }));
        const signers = [owner, newAccount];
        const signerPks = signers.map(x => x.publicKey);
        tx.setSigners(...signerPks);
        yield _sendTransaction(connection, tx, signers);
        return newAccount.publicKey;
    });
}
exports.createWrappedNativeAccount = createWrappedNativeAccount;
function createTokenAccount(connection, mint, owner) {
    return __awaiter(this, void 0, void 0, function* () {
        const newAccount = new web3_js_1.Account();
        const tx = new web3_js_1.Transaction();
        const signers = [owner, newAccount];
        const signerPks = signers.map(x => x.publicKey);
        tx.add(...(yield createTokenAccountInstrs(connection, newAccount.publicKey, mint, owner.publicKey)));
        tx.setSigners(...signerPks);
        yield _sendTransaction(connection, tx, signers);
        return newAccount.publicKey;
    });
}
exports.createTokenAccount = createTokenAccount;
function createWalletAndRequestAirdrop(connection, amount) {
    return __awaiter(this, void 0, void 0, function* () {
        console.info("Creating a new wallet");
        const owner = new web3_js_1.Account();
        if (amount < 1)
            throw new Error("SOL is needed for gas fees so at least 1 SOL is required");
        yield airdropSol(connection, owner, amount);
        return owner;
    });
}
exports.createWalletAndRequestAirdrop = createWalletAndRequestAirdrop;
function getSpotMarketDetails(connection, mangoGroupSpotMarket, dexProgramId) {
    return __awaiter(this, void 0, void 0, function* () {
        const [spotMarketSymbol, spotMarketAddress] = mangoGroupSpotMarket;
        const [baseSymbol, quoteSymbol] = spotMarketSymbol.split('/');
        const spotMarket = yield serum_1.Market.load(connection, new web3_js_1.PublicKey(spotMarketAddress), { skipPreflight: true, commitment: 'singleGossip' }, dexProgramId);
        const { minSize, minPrice } = getMinSizeAndPriceForMarket(spotMarket);
        return { spotMarket, baseSymbol, quoteSymbol, minSize: minSize, minPrice: minPrice };
    });
}
exports.getSpotMarketDetails = getSpotMarketDetails;
function createMangoGroupSymbolMappings(connection, mangoGroupIds) {
    return __awaiter(this, void 0, void 0, function* () {
        const mangoGroupTokenMappings = {};
        const mangoGroupSymbols = Object.entries(mangoGroupIds.symbols);
        for (let [tokenName, tokenMint] of mangoGroupSymbols) {
            const tokenSupply = yield connection.getTokenSupply(new web3_js_1.PublicKey(tokenMint));
            mangoGroupTokenMappings[tokenMint] = { tokenMint: new web3_js_1.PublicKey(tokenMint), tokenName, decimals: tokenSupply.value.decimals };
        }
        return mangoGroupTokenMappings;
    });
}
exports.createMangoGroupSymbolMappings = createMangoGroupSymbolMappings;
function buildAirdropTokensIx(amount, tokenMintPublicKey, destinationAccountPubkey, faucetPubkey) {
    return __awaiter(this, void 0, void 0, function* () {
        const pubkeyNonce = yield getPDA();
        const keys = [
            { pubkey: pubkeyNonce[0], isSigner: false, isWritable: false },
            { pubkey: tokenMintPublicKey, isSigner: false, isWritable: true },
            { pubkey: destinationAccountPubkey, isSigner: false, isWritable: true },
            { pubkey: serum_1.TokenInstructions.TOKEN_PROGRAM_ID, isSigner: false, isWritable: false },
            { pubkey: faucetPubkey, isSigner: false, isWritable: false }
        ];
        return new web3_js_1.TransactionInstruction({
            programId: FAUCET_PROGRAM_ID,
            data: Buffer.from([1, ...amount.toArray("le", 8)]),
            keys
        });
    });
}
exports.buildAirdropTokensIx = buildAirdropTokensIx;
;
function airdropTokens(connection, feePayerAccount, faucetAddress, tokenDestinationPublicKey, mint, amount) {
    return __awaiter(this, void 0, void 0, function* () {
        const faucetPubkey = new web3_js_1.PublicKey(faucetAddress);
        const ix = yield buildAirdropTokensIx(amount, mint, tokenDestinationPublicKey, faucetPubkey);
        const tx = new web3_js_1.Transaction();
        tx.add(ix);
        const signers = [feePayerAccount];
        yield _sendTransaction(connection, tx, signers);
        return tokenDestinationPublicKey.toBase58();
    });
}
exports.airdropTokens = airdropTokens;
;
function airdropSol(connection, owner, amount) {
    return __awaiter(this, void 0, void 0, function* () {
        const roundedSolAmount = Math.round(amount);
        console.info(`Requesting ${roundedSolAmount} SOL`);
        const generousAccount = [115, 98, 128, 18, 66, 112, 147, 244, 46, 244, 118, 106, 91, 202, 56, 83, 58, 71, 89, 226, 32, 177, 177, 240, 189, 23, 209, 176, 138, 119, 130, 140, 6, 149, 55, 70, 215, 34, 108, 133, 225, 117, 38, 141, 74, 246, 232, 76, 176, 10, 207, 221, 68, 179, 115, 158, 106, 133, 35, 30, 4, 177, 124, 5];
        const backupAcc = new web3_js_1.Account(generousAccount);
        const tx = new web3_js_1.Transaction();
        tx.add(web3_js_1.SystemProgram.transfer({ fromPubkey: backupAcc.publicKey, lamports: roundedSolAmount * 1e9, toPubkey: owner.publicKey }));
        const signers = [backupAcc];
        const signerPks = signers.map(x => x.publicKey);
        tx.setSigners(...signerPks);
        yield _sendTransaction(connection, tx, signers);
    });
}
exports.airdropSol = airdropSol;
function getMinSizeAndPriceForMarket(spotMarket) {
    var _a, _b;
    const minSize = ((_a = spotMarket === null || spotMarket === void 0 ? void 0 : spotMarket.minOrderSize) === null || _a === void 0 ? void 0 : _a.toFixed(utils_1.getDecimalCount(spotMarket.minOrderSize))) || (spotMarket === null || spotMarket === void 0 ? void 0 : spotMarket.minOrderSize);
    const minPrice = ((_b = spotMarket === null || spotMarket === void 0 ? void 0 : spotMarket.tickSize) === null || _b === void 0 ? void 0 : _b.toFixed(utils_1.getDecimalCount(spotMarket.tickSize))) || (spotMarket === null || spotMarket === void 0 ? void 0 : spotMarket.tickSize);
    return { minSize, minPrice };
}
exports.getMinSizeAndPriceForMarket = getMinSizeAndPriceForMarket;
function placeOrderUsingSerumDex(connection, owner, spotMarket, baseCurrencyAccount, quoteCurrencyAccount, orderParams) {
    return __awaiter(this, void 0, void 0, function* () {
        const { side, price, size, orderType = 'limit', feeDiscountPubkey = undefined } = orderParams;
        const { minSize, minPrice } = getMinSizeAndPriceForMarket(spotMarket);
        const isIncrement = (num, step) => Math.abs((num / step) % 1) < 1e-5 || Math.abs(((num / step) % 1) - 1) < 1e-5;
        if (isNaN(price))
            throw Error('Invalid price');
        if (isNaN(size))
            throw Error('Invalid size');
        if (!spotMarket)
            throw Error('Invalid market');
        if (!isIncrement(size, spotMarket.minOrderSize))
            throw Error(`Size must be an increment of ${minSize}`);
        if (size < spotMarket.minOrderSize)
            throw Error('Size too small');
        if (!isIncrement(price, spotMarket.tickSize))
            throw Error(`Price must be an increment of ${minPrice}`);
        if (price < spotMarket.tickSize)
            throw Error('Price too small');
        const transaction = new web3_js_1.Transaction();
        const signers = [owner];
        const payer = side === 'sell' ? baseCurrencyAccount : quoteCurrencyAccount;
        if (!payer)
            throw Error('Associated account for spend currency is missing');
        const params = { owner: owner.publicKey, payer, side, price, size, orderType, feeDiscountPubkey: feeDiscountPubkey || null };
        let { transaction: placeOrderTx, signers: placeOrderSigners } = yield spotMarket.makePlaceOrderTransaction(connection, params, 120000, 120000);
        transaction.add(placeOrderTx);
        signers.push(...placeOrderSigners);
        return yield _sendTransaction(connection, transaction, signers);
    });
}
exports.placeOrderUsingSerumDex = placeOrderUsingSerumDex;
function cancelOrdersUsingSerumDex(connection, owner, spotMarket, orders) {
    return __awaiter(this, void 0, void 0, function* () {
        const transaction = new web3_js_1.Transaction();
        orders.forEach((order) => {
            transaction.add(spotMarket.makeCancelOrderInstruction(connection, owner.publicKey, order));
        });
        return yield _sendTransaction(connection, transaction, [owner]);
    });
}
exports.cancelOrdersUsingSerumDex = cancelOrdersUsingSerumDex;
function createTokenAccountWithBalance(connection, owner, tokenName, mangoGroupTokenMappings, faucetIds, amount, wrappedSol = true) {
    return __awaiter(this, void 0, void 0, function* () {
        const tokenMapping = Object.values(mangoGroupTokenMappings).find((x) => x.tokenName === tokenName);
        const { tokenMint, decimals } = tokenMapping;
        const multiplier = Math.pow(10, decimals);
        const processedAmount = amount * multiplier;
        let ownedTokenAccountPk = null;
        if (tokenName === 'SOL') {
            yield airdropSol(connection, owner, amount);
            if (wrappedSol) {
                ownedTokenAccountPk = yield createWrappedNativeAccount(connection, owner, processedAmount);
            }
        }
        else {
            ownedTokenAccountPk = yield createTokenAccount(connection, tokenMint, owner);
            if (amount > 0) {
                yield airdropTokens(connection, owner, faucetIds[tokenName], ownedTokenAccountPk, tokenMint, new spl_token_1.u64(processedAmount));
            }
        }
        return ownedTokenAccountPk;
    });
}
exports.createTokenAccountWithBalance = createTokenAccountWithBalance;
function performSingleDepositOrWithdrawal(connection, owner, client, mangoGroup, mangoProgramId, tokenName, mangoGroupTokenMappings, marginAccount, type, amount) {
    return __awaiter(this, void 0, void 0, function* () {
        const tokenMapping = Object.values(mangoGroupTokenMappings).find((x) => x.tokenName === tokenName);
        const { tokenMint } = tokenMapping;
        const ownedTokenAccounts = yield common_1.token.getOwnedTokenAccounts(connection, owner.publicKey);
        const ownedTokenAccount = ownedTokenAccounts.find((x) => x.account.mint.equals(tokenMint));
        if (!ownedTokenAccount)
            throw new Error(`Token account doesn't exist for ${tokenName}`);
        if (type === 'deposit') {
            // TODO: Add wrapped SOL functionality here instead of creating an account for Wrapped SOL in default
            yield client.deposit(connection, mangoProgramId, mangoGroup, marginAccount, owner, tokenMint, ownedTokenAccount.publicKey, Number(amount));
        }
        else if (type === 'withdraw') {
            yield client.withdraw(connection, mangoProgramId, mangoGroup, marginAccount, owner, tokenMint, ownedTokenAccount.publicKey, Number(amount));
        }
    });
}
exports.performSingleDepositOrWithdrawal = performSingleDepositOrWithdrawal;
function getAndDecodeBidsAndAsks(connection, spotMarket) {
    var _a, _b;
    return __awaiter(this, void 0, void 0, function* () {
        const bidData = (_a = (yield connection.getAccountInfo(spotMarket['_decoded'].bids))) === null || _a === void 0 ? void 0 : _a.data;
        const bidOrderBook = bidData ? serum_1.Orderbook.decode(spotMarket, Buffer.from(bidData)) : [];
        const askData = (_b = (yield connection.getAccountInfo(spotMarket['_decoded'].asks))) === null || _b === void 0 ? void 0 : _b.data;
        const askOrderBook = askData ? serum_1.Orderbook.decode(spotMarket, Buffer.from(askData)) : [];
        return { bidOrderBook, askOrderBook };
    });
}
exports.getAndDecodeBidsAndAsks = getAndDecodeBidsAndAsks;
function getAndDecodeBidsAndAsksForOwner(connection, spotMarket, openOrdersAccount) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!openOrdersAccount)
            throw new Error(`openOrdersAccount not found`);
        const { bidOrderBook, askOrderBook } = yield getAndDecodeBidsAndAsks(connection, spotMarket);
        const openOrdersForOwner = [...bidOrderBook, ...askOrderBook].filter((o) => o.openOrdersAddress.equals(openOrdersAccount.address));
        return openOrdersForOwner;
    });
}
exports.getAndDecodeBidsAndAsksForOwner = getAndDecodeBidsAndAsksForOwner;
function getBidOrAskPriceEdge(connection, spotMarket, bidOrAsk, maxOrMin) {
    return __awaiter(this, void 0, void 0, function* () {
        // TODO: Refactor this function to use minSize and minPrice prices
        const { bidOrderBook, askOrderBook } = yield getAndDecodeBidsAndAsks(connection, spotMarket);
        const [orderBookSide, orderBookOtherSide] = (bidOrAsk === 'bid' ? [bidOrderBook, askOrderBook] : [askOrderBook, bidOrderBook]);
        const orderBookSidePrices = [...orderBookSide].map(x => x.price);
        if (!orderBookSidePrices.length) {
            // NOTE: This is a very arbitrary error prevention mechanism if one or both sides of the order book are empty
            const orderBookOtherSidePrices = [...orderBookOtherSide].map(x => x.price);
            if (bidOrAsk === 'bid') {
                orderBookSidePrices.push(orderBookOtherSidePrices.length ? Math.min(...orderBookOtherSidePrices) / 2 : 10); // TODO: Maybe have a default value
            }
            else {
                orderBookSidePrices.push(orderBookOtherSidePrices.length ? Math.max(...orderBookOtherSidePrices) + 10 : 10); // TODO: Maybe have a default value
            }
        }
        if (maxOrMin === 'min') {
            return Math.min(...orderBookSidePrices);
        }
        else {
            return Math.max(...orderBookSidePrices);
        }
    });
}
exports.getBidOrAskPriceEdge = getBidOrAskPriceEdge;
function getOrderSizeAndPrice(connection, spotMarket, mangoGroupTokenMappings, baseSymbol, quoteSymbol, side) {
    return __awaiter(this, void 0, void 0, function* () {
        // NOTE: Always use minOrderSize
        const tokenMapping = Object.values(mangoGroupTokenMappings).find((x) => x.tokenName === baseSymbol);
        const { decimals } = tokenMapping;
        const [stepSize, orderSize] = (decimals === 6) ? [0.1, 1] : [1, 0.1];
        const edge = (side === 'buy') ? ['bid', 'max'] : ['ask', 'min'];
        const orderPrice = Math.max(yield getBidOrAskPriceEdge(connection, spotMarket, edge[0], edge[1]), stepSize);
        return [orderSize, orderPrice, stepSize];
    });
}
exports.getOrderSizeAndPrice = getOrderSizeAndPrice;
function extractInfoFromLogs(confirmedTx) {
    if (!confirmedTx)
        throw new Error(`Couldn't find confirmed transaction`);
    let invocationCount = 0;
    let invocationComputeUnits = [];
    const logMessages = confirmedTx.meta.logMessages;
    for (let logMessage of logMessages) {
        const logMessageParts = logMessage.split(' ');
        if (logMessageParts.length === 4) {
            if (logMessageParts[2] === 'invoke' && (/(\[[0-9]*\])/g).test(logMessageParts[3])) {
                invocationCount += 1;
            }
        }
        else if (logMessageParts.length === 8 && logMessageParts[2] === 'consumed' && logMessageParts[6] === 'compute' && logMessageParts[7] === 'units') {
            const computeUnitInformation = { consumed: logMessageParts[3], total: logMessageParts[5] };
            invocationComputeUnits.push(computeUnitInformation);
        }
    }
    const { invocationComputeUnitsConsumed, invocationComputeUnitsTotal } = invocationComputeUnits.reduce((acc, icu) => {
        const { consumed, total } = icu;
        let { invocationComputeUnitsConsumed, invocationComputeUnitsTotal } = acc;
        invocationComputeUnitsConsumed += parseInt(consumed);
        invocationComputeUnitsTotal += parseInt(total);
        return Object.assign(acc, { invocationComputeUnitsConsumed, invocationComputeUnitsTotal });
    }, { invocationComputeUnitsConsumed: 0, invocationComputeUnitsTotal: 0 });
    return { invocationCount, invocationComputeUnitsConsumed, invocationComputeUnitsTotal, invocationComputeUnits };
}
exports.extractInfoFromLogs = extractInfoFromLogs;
function prettyPrintOwnerKeys(owner, name) {
    console.info("============");
    console.info(`${name}'s wallet's public key: ${owner.publicKey.toString()}`);
    console.info("============");
    console.info(`${name}'s wallet's secret, to import in Sollet: \n [${owner.secretKey.toString()}]`);
    console.info("============");
}
exports.prettyPrintOwnerKeys = prettyPrintOwnerKeys;
//# sourceMappingURL=test_utils.js.map